@@ -14,6 +14,7 @@ import fnmatch
 import subprocess
 import functools
 import re
+import json
 
 import apt_pkg
 
@@ -21,6 +22,63 @@ from UbuntuDrivers import kerneldetection
 
 system_architecture = ''
 lookup_cache = {}
+custom_supported_gpus_json = '/etc/custom_supported_gpus.json'
+
+
+class NvidiaPkgNameInfo(object):
+    '''Class to process NVIDIA package names'''
+    def __init__(self, pkg_name):
+        self._pkg_name = pkg_name
+        self._obsolete_name_scheme = False
+        self._server = False
+        self._open = False
+        self._major_ver = -1
+        self._flavour = ''
+        self.is_valid = False
+        self._process_name(self._pkg_name)
+
+    def _process_name(self, name):
+        if 'nvidia' not in name:
+            logging.debug('NvidiaPkgNameInfo: %s is not an NVIDIA package. Skipping', name)
+            return
+
+        pattern = re.compile('nvidia-([0-9]+)')
+        match = pattern.match(name)
+
+        # Obsolete naming such as nvidia-340
+        if match:
+            self._obsolete_name_scheme = True
+            self._major_ver = match.group(1)
+            self._flavour = self._major_ver
+            self.is_valid = True
+            return
+
+        # Recent naming such as nvidia-driver-525
+        pattern = re.compile('nvidia-driver-([0-9]+)(.*)')
+        match = pattern.match(name)
+
+        if match:
+            self._server = match.group(0).find('-server') != -1
+            self._open = match.group(0).find('-open') != -1
+            self._flavour = '%s%s%s' % (match.group(1),
+                                        '-server' if self._server else '',
+                                        '-open' if self._open else '')
+            self.is_valid = True
+
+    def has_obsolete_name_scheme(self):
+        return self._obsolete_name_scheme
+
+    def is_server(self):
+        return self._server
+
+    def is_open(self):
+        return self._open
+
+    def get_major_version(self):
+        return self._major_ver
+
+    def get_flavour(self):
+        return self._flavour
 
 
 def get_apt_arch():
@@ -190,6 +248,37 @@ def _apt_cache_modalias_map(apt_cache):
     return result2
 
 
+def path_get_custom_supported_gpus():
+    return custom_supported_gpus_json
+
+
+def package_get_nv_allowing_driver(did):
+    '''Get nvidia allowing driver for specific devices.
+
+    did: 0x1234
+    Return the situable nvidia driver version for it.
+    '''
+    path = path_get_custom_supported_gpus()
+    version = None
+    try:
+        with open(path, "r") as stream:
+            try:
+                gpus = list(json.load(stream)['chips'])
+                for gpu in gpus:
+                    if gpu['devid'] == did:
+                        version = gpu['branch'].split('.')[0]
+                        logging.info("Found a specific nv driver version %s for %s(%s)" %
+                                     (version, gpu['name'], did))
+                        break
+            except Exception:
+                logging.debug('package_get_nv_allowing_driver(): unexpected json detected.')
+                pass
+    except Exception:
+        logging.debug('package_get_nv_allowing_driver(): unable to read %s' % path)
+        pass
+    return version
+
+
 def packages_for_modalias(apt_cache, modalias):
     '''Search packages which match the given modalias.
 
@@ -207,6 +296,20 @@ def packages_for_modalias(apt_cache, modalias):
     pat, bus_map = cache_map.get(modalias.split(':', 1)[0], (None, {}))
     if pat is None or not pat.match(modalias):
         return []
+    vid, did = _get_vendor_model_from_alias(modalias)
+    nvamd = None
+    found = 0
+    if vid == "10DE":
+        nvamd = package_get_nv_allowing_driver("0x" + did)
+        nvamdn = "nvidia-driver-%s" % nvamd
+        nvamda = "pci:v000010DEd0000%s*" % did
+        for p in apt_cache.packages:
+            if p.get_fullname().split(':')[0] == nvamdn:
+                bus_map[nvamda] = set([nvamdn])
+                found = 1
+        if nvamd is not None and not found:
+            logging.debug('%s is not in the package pool.' % nvamdn)
+
     for alias in bus_map:
         if fnmatch.fnmatchcase(modalias.lower(), alias.lower()):
             for p in bus_map[alias]:
@@ -303,10 +406,46 @@ def _pkg_get_support(apt_cache, pkg):
     return support
 
 
-def _is_runtimepm_supported(pkg, alias):
+def _is_nv_allowing_runtimepm_supported(alias, ver):
+    '''alias: e.g. pci:v000010DEd000024BAsv0000103Csd000089C6bc03sc00i00'''
+    result = re.search('pci:v0000(.*)d0000(.*)sv(.*)', alias)
+    vid = result.group(1)
+    did = result.group(2)
+    if vid != "10DE":
+        return False
+    did = "0x%s" % did
+    path = path_get_custom_supported_gpus()
+    try:
+        with open(path, "r") as stream:
+            try:
+                gpus = list(json.load(stream)['chips'])
+                for gpu in gpus:
+                    if gpu['devid'] == did and 'runtimepm' in gpu['features']:
+                        if gpu['branch'].split('.')[0] != ver:
+                            logging.debug('Candidate version does not match %s != %s' %
+                                          (gpu['branch'].split('.')[0], ver))
+                            return False
+                        logging.info("Found runtimepm supports on %s." % did)
+                        return True
+            except Exception:
+                logging.debug('_is_nv_allowing_runtimepm_supported(): unexpected json detected')
+                pass
+    except Exception:
+        logging.debug('_is_nv_allowing_runtimepm_supported(): unable to read %s' % path)
+        pass
+    return False
+
+
+def _is_runtimepm_supported(apt_cache, pkg, alias):
     '''Check if the package supports runtimepm for the given modalias'''
     try:
-        m = pkg.candidate.record['PmAliases']
+        depcache = apt_pkg.DepCache(apt_cache)
+        candidate = depcache.get_candidate_ver(pkg)
+        records = apt_pkg.PackageRecords(apt_cache)
+        records.lookup(candidate.file_list[0])
+        section = apt_pkg.TagSection(records.record)
+        ver = candidate.ver_str.split('.')[0]
+        m = section['PmAliases']
     except (KeyError, AttributeError, UnicodeDecodeError):
         return False
     else:
@@ -315,6 +454,8 @@ def _is_runtimepm_supported(pkg, alias):
 
         n = m[m.find('(')+1: m.find(')')]
         modaliases = n.split(', ')
+        if _is_nv_allowing_runtimepm_supported(alias, ver):
+            return True
         return any(fnmatch.fnmatch(alias.lower(), regex.lower()) for regex in modaliases)
 
 
@@ -456,7 +597,7 @@ def system_driver_packages(apt_cache=None, sys_path=None, freeonly=False, includ
                     'free': _is_package_free(apt_cache, p),
                     'from_distro': _is_package_from_distro(apt_cache, p),
                     'support': _pkg_get_support(apt_cache, p),
-                    'runtimepm': _is_runtimepm_supported(p, alias)
+                    'runtimepm': _is_runtimepm_supported(apt_cache, p, alias)
                 }
             (vendor, model) = _get_db_name(syspath, alias)
             if vendor is not None:
@@ -515,19 +656,16 @@ def _get_headless_no_dkms_metapackage(pkg, apt_cache):
     depcache = apt_pkg.DepCache(apt_cache)
     name = pkg.name
 
-    pattern = re.compile('nvidia-([0-9]+)')
-    match = pattern.match(name)
-    if match:
-        logging.debug('Legacy driver detected: %s. Skipping.', name)
+    nvidia_info = NvidiaPkgNameInfo(name)
+    if not nvidia_info.is_valid:
+        logging.debug('Unsupported driver detected: %s. Skipping' % name)
         return metapackage
 
-    pattern = re.compile('nvidia-driver-([0-9]+)(.*)')
-    match = pattern.match(name)
-    if not match:
-        logging.debug('No flavour can be found in %s. Skipping.', name)
+    if nvidia_info.has_obsolete_name_scheme():
+        logging.debug('Legacy driver detected: %s. Skipping.' % name)
         return metapackage
 
-    candidate_flavour = '%s%s' % (match.group(1), match.group(2))
+    candidate_flavour = nvidia_info.get_flavour()
     candidate = 'nvidia-headless-no-dkms-%s' % (candidate_flavour)
 
     try:
@@ -826,22 +964,13 @@ def get_desktop_package_list(apt_cache, sys_path=None, free_only=False, include_
 
 def nvidia_desktop_pre_installation_hook(to_install):
     '''Applies changes that need to happen before installing the NVIDIA drivers'''
-    with_nvidia_kms = False
 
     # Enable KMS if nvidia >= 470
     for package_name in to_install:
-        pattern = re.compile('nvidia-driver-([0-9]+)(.*)')
-        match = pattern.match(package_name)
-        if match:
-            try:
-                version = int(match.group(1))
-            except ValueError:
-                pass
-            finally:
-                with_nvidia_kms = version >= 470
-
-    if with_nvidia_kms:
-        set_nvidia_kms(1)
+        nvidia_info = NvidiaPkgNameInfo(package_name)
+        if nvidia_info.get_major_version() >= 470:
+            set_nvidia_kms(1)
+            return
 
 
 def nvidia_desktop_post_installation_hook():
@@ -1166,10 +1295,16 @@ def _cmp_gfx_alternatives(x, y):
 def _cmp_gfx_alternatives_gpgpu(x, y):
     '''Compare two graphics driver names in terms of preference. (server)
 
+    -open always sorts after non-open.
     -server always sorts before non-server.
     LTSB (Long Term Support Branch) always sorts before NFB (New Feature Branch).
     Legacy always sorts before Beta.
     '''
+    if x.endswith('-open') and not y.endswith('-open'):
+        return -1
+    if not x.endswith('-open') and y.endswith('-open'):
+        return 1
+
     if x.endswith('-server') and not y.endswith('-server'):
         return 1
     if not x.endswith('-server') and y.endswith('-server'):
@@ -1286,14 +1421,13 @@ def get_linux_modules_metapackage(apt_cache, candidate):
 
     depcache = apt_pkg.DepCache(apt_cache)
 
-    if 'nvidia' not in candidate:
-        logging.debug('Non NVIDIA linux-modules packages are not supported at this time: %s. Skipping', candidate)
+    nvidia_info = NvidiaPkgNameInfo(candidate)
+    if not nvidia_info.is_valid:
+        logging.debug('Non NVIDIA linux-modules packages are not supported at this time: %s. Skipping' % candidate)
         return metapackage
 
-    pattern = re.compile('nvidia-([0-9]+)')
-    match = pattern.match(candidate)
-    if match:
-        logging.debug('Legacy driver detected: %s. Skipping.', candidate)
+    if nvidia_info.has_obsolete_name_scheme():
+        logging.debug('Legacy driver detected: %s. Skipping.' % candidate)
         return metapackage
 
     linux_image_meta = get_linux_image(apt_cache)
@@ -1303,16 +1437,10 @@ def get_linux_modules_metapackage(apt_cache, candidate):
     if linux_image:
         linux_flavour = linux_image.replace('linux-image-', '')
     else:
-        logging.debug('No linux-image can be found for %s. Skipping.', candidate)
+        logging.debug('No linux-image can be found for %s. Skipping.' % candidate)
         return metapackage
 
-    pattern = re.compile('nvidia-.*-([0-9]+)(.*)')
-    match = pattern.match(candidate)
-    if not match:
-        logging.debug('No flavour can be found in %s. Skipping.', candidate)
-        return metapackage
-
-    candidate_flavour = '%s%s' % (match.group(1), match.group(2))
+    candidate_flavour = nvidia_info.get_flavour()
     linux_modules_candidate = 'linux-modules-nvidia-%s-%s' % (candidate_flavour, linux_flavour)
 
     try:
